"""
This file contains number of functions to handle different software versions compatibility
"""
import json

from django.db.models import Model, Field, BigIntegerField, IntegerField
from typing import Dict, Any, Optional, Union, Tuple, List, Type

import django
from django.db import connection, connections, models, DefaultConnectionProxy, migrations


def zip_longest(*args, **kwargs):
    """
    https://docs.python.org/3.5/library/itertools.html#itertools.zip_longest
    """
    try:
        # python 3
        from itertools import zip_longest
        return zip_longest(*args, **kwargs)
    except ImportError:
        # python 2.7
        from itertools import izip_longest
        return izip_longest(*args, **kwargs)


def jsonb_available():  # type: () -> bool
    """
    Checks if we can use JSONField.
    It is available since django 1.9 and doesn't support Postgres < 9.4
    :return: Bool
    """
    return get_postgres_version() >= (9, 4) and django.VERSION >= (1, 9)


def hstore_available():  # type: () -> bool
    """
    Checks if we can use HStoreField.
    It is available since django 1.8
    :return: Bool
    """
    return django.VERSION >= (1, 8)


def array_available():  # type: () -> bool
    """
    Checks if we can use ArrayField.
    It is available since django 1.8
    :return: Bool
    """
    return django.VERSION >= (1, 8)


def hstore_serialize(value):  # type: (Dict[Any, Any]) -> Dict[str, str]
    """
    Django before 1.10 doesn't convert HStoreField values to string automatically
    Which causes a bug in cursor.execute(). This function converts all key/values to string
    :param value: A dict
    :return:
    """
    val = {
        str(k): json.dumps(v) if isinstance(v, (dict, list)) else str(v)
        for k, v in value.items()
    }
    return val


def get_postgres_version(using=None, as_tuple=True):
    # type: (Optional[str], bool) -> Union[Tuple[int], int]
    """
    Returns Postgres server version used
    :param using: Connection alias to use
    :param as_tuple: If true, returns result as tuple, otherwise as concatenated integer
    :return: Database version as tuple (major, minor, revision) if as_tuple is true.
        A single number major*10000 + minor*100 + revision if false.
    """
    conn = connection if using is None else connections[using]
    num = conn.cursor().connection.server_version
    return (int(num / 10000), int(num % 10000 / 100), num % 100) if as_tuple else num


def get_field_db_type(field, conn):
    # type: (models.Field, DefaultConnectionProxy) -> str
    """
    Get database field type used for this field.
    :param field: django.db.models.Field instance
    :param conn: Database connection used
    :return: Database type name (str)
    """
    # db_type() as id field returned 'serial' instead of 'integer' here
    # rel_db_type() return integer, but it is not available before django 1.10
    # rel_db_type() returns serial[] for arrays.
    # cast_db_type() is not well documented and returns serial too.
    db_type = field.db_type(conn)

    # Some external libraries may add column fields here
    # Let's cut them
    cut_phrases = [
        'CONSTRAINT', 'NOT NULL', 'NULL', 'CHECK', 'DEFAULT', 'UNIQUE', 'PRIMARY KEY', 'REFERENCES', 'COLLATE'
    ]
    for ph in cut_phrases:
        db_type = db_type.split(ph, 1)[0]

    db_type = db_type.strip()

    if 'bigserial' in db_type:
        db_type = db_type.replace('bigserial', BigIntegerField().db_type(conn))
    elif 'serial' in db_type:
        db_type = db_type.replace('serial', IntegerField().db_type(conn))

    return db_type


def get_model_fields(model, concrete_only=False):
    # type: (Type[Model], bool) -> List[Field]
    """
    Returns all model fields.
    :param model: Model to get fields for
    :param concrete_only: If True, returns only fields which have database columns
    :return: A list of fields
    """
    if hasattr(model._meta, 'get_fields'):
        # Django 1.8+
        fields = model._meta.get_fields()
        return [f for f in fields if f.concrete] if concrete_only else fields
    else:
        # Django 1.7
        if concrete_only:
            return [f[0] for f in model._meta.get_concrete_fields_with_model()]
        else:
            return [f[0] for f in model._meta.get_fields_with_model()]


# Postgres 9.4 has JSONB support, but doesn't support concat operator (||)
# So I've taken function to solve the problem from
# https://stackoverflow.com/questions/30101603/merging-concatenating-jsonb-columns-in-query
class Postgres94MergeJSONBMigration(migrations.RunSQL):
    FUNCTION_NAME = "django_pg_bulk_update_jsonb_merge"

    SQL_TEMPLATE = """
    CREATE OR REPLACE FUNCTION %s(jsonb1 JSONB, jsonb2 JSONB)
        RETURNS JSONB AS $$
        DECLARE
          result JSONB;
          v RECORD;
        BEGIN
           result = (
             SELECT json_object_agg(KEY,value)
             FROM (
               SELECT jsonb_object_keys(jsonb1) AS KEY,
                 1::INT AS jsb,
                 jsonb1 -> jsonb_object_keys(jsonb1) AS value
               UNION SELECT jsonb_object_keys(jsonb2) AS KEY,
                 2::INT AS jsb,
                 jsonb2 -> jsonb_object_keys(jsonb2) AS value ) AS t1
            );
                         
            RETURN COALESCE(result, '{}'::JSONB);
        END;
        $$ LANGUAGE plpgsql;
    """

    REVERSE_SQL_TEMPLATE = """
    DROP FUNCTION IF EXISTS %s(jsonb1 JSONB, jsonb2 JSONB);
    """

    def __init__(self, **kwargs):
        sql = self.SQL_TEMPLATE % self.FUNCTION_NAME
        kwargs['reverse_sql'] = self.REVERSE_SQL_TEMPLATE % self.FUNCTION_NAME
        super(Postgres94MergeJSONBMigration, self).__init__(sql, **kwargs)



"""
This file contains django manager and a mixin to integrate library with your models

Examples:
    from django.db import models
    from django_pg_bulk_update import

    # Simple manager
    class TestModel(models.Model):
        objects = BulkUpdateManager()

    # Custom manager
    class CustomManager(models.Manager, BulkUpdateManagerMixin):
        pass  # Your functionality here

    class TestModel(models.Model):
        objects = CustomManager()
"""
from django.db import models
from typing import Tuple, Optional

from .types import TUpdateValues, TFieldNames, TSetFunctions, TOperators
from .query import bulk_update, bulk_update_or_create

class BulkUpdateQuerySet(models.QuerySet):
    def bulk_create(self, objs, batch_size=None, conflict_target=None, conflict_action=None):
        pass


class BulkUpdateManagerMixin:
    """
    A mixin, adding bulk updates methods to any django manager
    It automatically fetches using and model parameters from manager.
    You can set database alias to use directly by db_manager() method
    """
    def bulk_update(self, values, key_fields='id', set_functions=None, key_fields_ops=(), batch_size=None,
                    batch_delay=0):
        # type: (TUpdateValues, TFieldNames, TSetFunctions, TOperators, Optional[int], float) -> int
        """
        Updates multiple records of a given model, finding them by key_fields.

        Example:
        # Test model
        class TestModel(models.Model):
            name = models.CharField(max_length=50)
            int_field = models.IntegerField(default=1)

        # Create test data
        TestModel.objects.bulk_create([TestModel(pk=i, name="item%d" % i) for i in range(1, 4)])

        # Call update. Does only 1 database query.
        updated = bulk_update(TestModel, {
            "item1": {
                "name": "updated1",
                "int_field": 1
            },
            "item2": {
                "name": "updated2",
                "int_field": 2
            }
        }, key_fields="name")

        print(updated)
        # Outputs: 2

        print(list(TestModel.objects.all().order_by("id").values("id", "name", "int_field")))
        # Outputs: [
        #     {"id": 1, "name": "updated1", "int_field": 2},
        #     {"id": 2, "name": "updated1", "int_field": 3},
        #     {"id": 3, "name": "item3", "int_field": 0}
        # ]

        :param values: Data to update. All items must update same fields!!!
            It can come in 2 forms:
            + Iterable of dicts. Each dict is update or create data. Each dict must contain all key_fields as keys.
                You can't update key_fields with this format.
            + Dict of key_values: update_fields_dict
                - key_values can be iterable or single object.
                - If iterable, key_values length must be equal to key_fields length.
                - If single object, key_fields is expected to have 1 element
        :param key_fields: Field names, by which items would be selected.
            It can be a string, if there's only one key field or iterable of strings for multiple keys
        :param set_functions: Functions to set values.
            Should be a dict of field name as key, function as value.
            Default function is eq.
            Functions: [eq, =; incr, +; concat, ||]
            Example: {'name': 'eq', 'int_fields': 'incr'}
        :param key_fields_ops: Key fields compare operators.
            It can be dict with field_name from key_fields as key, operation name as value
            Or an iterable of operations in key_fields order.
            The default operator is eq (it will be used for all fields, not set directly).
            Operators: [in; !in; gt, >; lt, <; gte, >=; lte, <=; !eq, <>, !=; eq, =, ==]
            Example: ('eq', 'in') or {'a': 'eq', 'b': 'in'}.
        :param batch_size: Optional. If given, data is split it into batches of given size.
            Each batch is queried independently.
        :param batch_delay: Delay in seconds between batches execution, if batch_size is not None.
        :return: Number of records updated
        """
        self._for_write = True
        using = self.db

        return bulk_update(self.model, values, key_fds=key_fields, using=using, set_functions=set_functions,
                           key_fields_ops=key_fields_ops, batch_size=batch_size, batch_delay=batch_delay)

    def bulk_update_or_create(self, values, key_fields='id', set_functions=None, update=True, key_is_unique=True,
                              batch_size=None, batch_delay=0):
        # type: (TUpdateValues, TFieldNames, TSetFunctions, bool, bool, Optional[int], float) -> int
        """
        Searches for records, given in values by key_fields. If records are found, updates them from values.
        If not found - creates them from values. Note, that all fields without default value must be present in values.

        :param values: Data to update. All items must update same fields!!!
            It can come in 2 forms:
            + Iterable of dicts. Each dict is update or create data. Each dict must contain all key_fields as keys.
                You can't update key_fields with this format.
            + Dict of key_values: update_fields_dict
                - key_values can be iterable or single object.
                - If iterable, key_values length must be equal to key_fields length.
                - If single object, key_fields is expected to have 1 element
        :param key_fields: Field names, by which items would be selected.
            It can be a string, if there's only one key field or iterable of strings for multiple keys
        :param set_functions: Functions to set values.
            Should be a dict of field name as key, function as value.
            Default function is eq.
            Functions: [eq, =; incr, +; concat, ||]
            Example: {'name': 'eq', 'int_fields': 'incr'}
        :param update: If this flag is not set, existing records will not be updated
        :param key_is_unique: Settings this flag to False forces library to use 3-query transactional update,
            not INSERT ... ON CONFLICT.
        :param batch_size: Optional. If given, data is split it into batches of given size.
            Each batch is queried independently.
        :param batch_delay: Delay in seconds between batches execution, if batch_size is not None.
        :return: Number of records created or updated
        """
        self._for_write = True
        using = self.db

        return bulk_update_or_create(self.model, values, key_fields=key_fields, using=using,
                                     set_functions=set_functions, update=update, key_is_unique=key_is_unique,
                                     batch_size=batch_size, batch_delay=batch_delay)


class BulkUpdateManager(models.Manager, BulkUpdateManagerMixin):
    """
    A manager, adding bulk update methods to any model
    """
    pass



from collections import defaultdict
from typing import Iterable, Union, Dict, Tuple, Any, Optional, Type

import six
from django.db.models import Model, Field

from .clause_operators import AbstractClauseOperator, EqualClauseOperator
from .set_functions import AbstractSetFunction, EqualSetFunction

TFieldNames = Union[str, Iterable[str]]

TOperator = Union[str, AbstractClauseOperator]
TOperatorsValid = Tuple['FieldDescriptor']

TOperators = Union[Dict[str, TOperator], Iterable[TOperator]]
TUpdateValuesValid = Dict[Tuple[Any], Dict[str, Any]]
TUpdateValues = Union[Union[TUpdateValuesValid, Dict[Any, Dict[str, Any]]], Iterable[Dict[str, Any]]]
TSetFunction = Union[str, AbstractSetFunction]
TSetFunctions = Optional[Dict[str, TSetFunction]]
TSetFunctionsValid = Tuple['FieldDescriptor']


class FieldDescriptor(object):
    """
    This class is added in order to make passing parameters in queries easier
    """
    __slots__ = ['name', '_set_function', '_key_operator', '_prefix']

    def __init__(self, name, set_function=None, key_operator=None):
        # type: (str, TSetFunction, TOperator) -> None
        self.name = name
        self.set_function = set_function
        self.key_operator = key_operator
        self._prefix = ''

    def get_field(self, model):
        # type: (Type[Model]) -> Field
        """
        Returns model field, described by this descriptor
        :param model: django.db.models.Model subclass
        :return: django.db.fields.Field instance
        """
        return model._meta.get_field(self.name)

    @property
    def set_function(self):
        # type: () -> AbstractSetFunction
        """
        Returns set_function for this field descriptor.
        :return: AbstractSetFunction instance
        """
        return self._set_function

    @set_function.setter
    def set_function(self, val):
        # type: (Union[None, str, AbstractSetFunction]) -> None
        """
        Changes set_function for this field_descriptor.
        :param val: Set function name or instance. Defaults to EqualSetFunction() if None is passed
        :return:
        """
        if val is None:
            self._set_function = EqualSetFunction()
        elif isinstance(val, six.string_types):
            self._set_function = AbstractSetFunction.get_function_by_name(val)()
        elif isinstance(val, AbstractSetFunction):
            self._set_function = val
        else:
            raise TypeError("Invalid set function type: %s" % str(type(val)))

    @property
    def key_operator(self):
        # type: () -> AbstractClauseOperator
        """
        Returns operator to use in comparison
        :return: AbstractKeyOperator instance
        """
        return self._key_operator

    @key_operator.setter
    def key_operator(self, val):
        # type: (Union[None, str, AbstractClauseOperator]) -> None
        """
        Sets comparison operator for this field descriptor
        :param val: String name of operator or AbstractClauseOperator instance.
            Defaults to EqualClauseOperator if None is passed
        :return: None
        """
        if val is None:
            self._key_operator = EqualClauseOperator()
        elif isinstance(val, six.string_types):
            self._key_operator = AbstractClauseOperator.get_operator_by_name(val)()
        elif isinstance(val, AbstractClauseOperator):
            self._key_operator = val
        else:
            raise TypeError("Invalid key operator type: %s" % str(type(val)))

    def set_prefix(self, prefix, index=None): # type: (str, Optional[int]) -> None
        """
        Sets prefix to use in values query part. It is used to divide key fields from update and default fields
        :param prefix: Prefix to use
        :param index: field can be used more than once in conditions. Set this index to prevent duplicates.
        :return:
        """
        self._prefix = prefix
        if index is not None:
            self._prefix += '_%d' % index

    @property
    def prefixed_name(self):  # type: () -> str
        """
        Returns prefixed name of the field
        :return:
        """
        if self._prefix is None:
            raise ValueError('prefix has not been set yet')
        return "%s__%s" % (self._prefix, self.name)



"""
This function contains operators used in WHERE query part
"""
from typing import Type, Optional, Any, Tuple, Iterable, Dict

from django.db import DefaultConnectionProxy
from django.db.models import Field

from .compatibility import array_available, get_field_db_type
from .utils import get_subclasses, format_field_value


class AbstractClauseOperator(object):
    inverse = False
    names = set()

    def get_django_filters(self, name, value):
        # type: (str, Any) -> Dict[str, Any]
        """
        This method should return parameter name to use in django QuerySet.filter() kwargs
        :param name: Name of the parameter
        :param value: Value of the parameter
        :return: kwargs to pass to Q() object constructor
        """
        raise NotImplementedError("%s must implement get_django_filter method" % self.__class__.__name__)

    @classmethod
    def get_operator_by_name(cls, name):  # type: (str) -> Optional[Type[AbstractClauseOperator]]
        """
        Finds subclass of AbstractOperation applicable to given operation name
        :param name: String name to search
        :return: AbstractOperation subclass if found, None instead
        """
        try:
            return next(sub_cls for sub_cls in get_subclasses(cls, recursive=True) if name in sub_cls.names)
        except StopIteration:
            raise ValueError("Operator with name '%s' doesn't exist" % name)

    def get_sql_operator(self):  # type: () -> str
        """
        If get_sql operator is simple binary operator like "field <op> val", this functions returns operator
        :return: str
        """
        raise NotImplementedError("%s must implement get_sql_operator method" % self.__class__.__name__)

    def get_sql(self, table_field, value):  # type: (str, str) -> str
        """
        This method returns SQL of this operator.
        :param table_field: Table field string, already quoted and formatted on which clause is used
        :param value: Value string, already quoted and formatted. Can contain function calls.
        :return: String
        """
        return "%s %s %s" % (table_field, self.get_sql_operator(), value)

    def format_field_value(self, field, val, connection, cast_type=False, **kwargs):
        # type: (Field, Any, DefaultConnectionProxy, bool, **Any) -> Tuple[str, Tuple[Any]]
        """
        Formats value, according to field rules
        :param field: Django field to take format from
        :param val: Value to format
        :param connection: Connection used to update data
        :param cast_type: Adds type casting to sql if flag is True
        :param kwargs: Additional arguments, if needed
        :return: A tuple: sql, replacing value in update and a tuple of parameters to pass to cursor
        """
        return format_field_value(field, val, connection, cast_type=cast_type)


class AbstractArrayValueOperator(AbstractClauseOperator):
    """
    Abstract class partial, that handles an array of field values as input
    """
    def format_field_value(self, field, val, connection, **kwargs):
        assert isinstance(val, Iterable), "'%s' value must be iterable" % self.__class__.__name__

        # With in operator we should pass array of values instead of single field value.
        # So let's validate it as Array of this field
        if array_available():
            from django.contrib.postgres.fields import ArrayField
            arr_field = ArrayField(field)
            arr_field.model = field.model
            return super(AbstractArrayValueOperator, self).format_field_value(arr_field, val, connection, **kwargs)
        else:
            # This means, we use django < 1.8. Try converting it manually
            db_type = get_field_db_type(field, connection)
            tpl = "ARRAY[%s]::%s[]"
            val = list(val)

            placeholders, values = [], []
            for item in val:
                p, v = super(AbstractArrayValueOperator, self).format_field_value(field, item, connection, **kwargs)
                placeholders.append(p)
                values.extend(v)

            query = tpl % (', '.join(placeholders), db_type)
            return query, values


class EqualClauseOperator(AbstractClauseOperator):
    names = {'eq', '=', '=='}

    def get_django_filters(self, name, value):
        return {name: value}

    def get_sql_operator(self):
        return '='


class NotEqualClauseOperator(EqualClauseOperator):
    names = {'!eq', '!=', '<>'}
    inverse = True

    def get_sql_operator(self):
        return '!='


class InClauseOperator(AbstractArrayValueOperator):
    names = {'in'}
    django_operation = 'in'

    def get_django_filters(self, name, value):
        return {'%s__in' % name: value}

    def get_sql_operator(self):  # type: () -> str
        raise NotImplementedError("%s implements get_sql method, this method shouldn't be called"
                                  % self.__class__.__name__)

    def get_sql(self, table_field, value):
        # We can't simply use in as format_field_value will return ARRAY, not set of records
        return '%s = ANY(%s)' % (table_field, value)


class NotInClauseOperation(InClauseOperator):
    names = {'!in'}
    inverse = True

    def get_sql_operator(self):
        return super(NotInClauseOperation, self).get_sql_operator()

    def get_sql(self, table_field, value):
        return 'NOT %s' % super(NotInClauseOperation, self).get_sql(table_field, value)


class LTClauseOperator(AbstractClauseOperator):
    names = {'lt', '<'}

    def get_django_filters(self, name, value):
        return {'%s__lt' % name: value}

    def get_sql_operator(self):
        return '<'


class GTClauseOperator(AbstractClauseOperator):
    names = {'gt', '>'}

    def get_django_filters(self, name, value):
        return {'%s__gt' % name: value}

    def get_sql_operator(self):
        return '>'


class GTEClauseOperator(AbstractClauseOperator):
    names = {'gte', '>='}

    def get_django_filters(self, name, value):
        return {'%s__gte' % name: value}

    def get_sql_operator(self):
        return '>='


class LTEClauseOperator(AbstractClauseOperator):
    names = {'lte', '<='}

    def get_django_filters(self, name, value):
        return {'%s__lte' % name: value}

    def get_sql_operator(self):
        return '<='


class BetweenClauseOperator(AbstractArrayValueOperator):
    names = {'between'}

    def get_django_filters(self, name, value):
        assert isinstance(value, Iterable) and len(value) == 2, "value must be iterable of size 2"
        return {'%s__gte' % name: value[0], '%s__lte' % name: value[1]}

    def get_sql_operator(self):  # type: () -> str
        raise NotImplementedError("%s implements get_sql method, this method shouldn't be called"
                                  % self.__class__.__name__)

    def get_sql(self, table_field, value):  # type: (str, str) -> str
        # Postgres enumerates arrays from 1
        return "%s BETWEEN %s[1] AND %s[2]" % (table_field, value, value)



"""
This file contains classes, describing functions which set values to fields.
"""
import datetime
from typing import Type, Optional, Any, Tuple, Dict

import pytz
from django.db import DefaultConnectionProxy
from django.db.models import Field, Model

from .compatibility import get_postgres_version, jsonb_available, Postgres94MergeJSONBMigration, hstore_serialize,\
    hstore_available
from .utils import get_subclasses, format_field_value

# When doing increment operations, we need to replace NULL values with something
# This dictionary contains field defaults by it's class name.
# I don't use classes as keys not to import them here
base_datetime = datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=pytz.utc)
NULL_DEFAULTS = {
    # Standard django types
    'IntegerField': 0,
    'BigIntegerField': 0,
    'SmallIntegerField': 0,
    'PositiveIntegerField': 0,
    'PositiveSmallIntegerField': 0,
    'FloatField': 0,
    'DecimalField': 0,

    'CharField': '',
    'TextField': '',
    'EmailField': '',
    'FilePathField': '',
    'SlugField': '',
    'CommaSeparatedIntegerField': '0,0',
    'URLField': '',
    'BinaryField': b'',
    'UUIDField': '',

    'DateField': base_datetime.date(),
    'DateTimeField': base_datetime,
    'DurationField': datetime.timedelta(seconds=0),
    'TimeField': base_datetime.time(),

    # This may be incorrect, but there is no any chance to understand, what is correct here
    'BooleanField': True,
    'NullBooleanField': True,

    # These fields can't be null, but I add them for compatibility
    'AutoField': 0,
    'BigAutoField': 0,

    # Postgres specific types
    'ArrayField': [],
    'CICharField': '',
    'CIEmailField': '',
    'CITextField': '',
    'HStoreField': {},
    'JSONField': {},

    'IntegerRangeField': (0, 0),
    'BigIntegerRangeField': (0, 0),
    'FloatRangeField': (0, 0),
    'DateTimeRangeField': (base_datetime, base_datetime),
    'DateRangeField': (base_datetime.date(), base_datetime.date())
}


# JSONField is available in django 1.9+ only
# I create fake class for previous version in order to just skip isinstance(item, JSONField) if branch
if jsonb_available():
    from django.contrib.postgres.fields import JSONField
else:
    class JSONField:
        pass


class AbstractSetFunction(object):
    names = set()

    # If set function supports any field class, this should be None.
    # Otherwise a set of class names supported
    supported_field_classes = None

    def format_field_value(self, field, val, connection, cast_type=False, **kwargs):
        # type: (Field, Any, DefaultConnectionProxy, bool, **Any) -> Tuple[str, Tuple[Any]]
        """
        Formats value, according to field rules
        :param field: Django field to take format from
        :param val: Value to format
        :param connection: Connection used to update data
        :param cast_type: Adds type casting to sql if flag is True
        :param kwargs: Additional arguments, if needed
        :return: A tuple: sql, replacing value in update and a tuple of parameters to pass to cursor
        """
        return format_field_value(field, val, connection, cast_type=cast_type)

    def modify_create_params(self, model, key, kwargs):
        # type: (Type[Model], str, Dict[str, Any]) -> Dict[str, Any]
        """
        This method modifies parameters before passing them to model(**kwargs)
        :param model: Model to get field from
        :param key: Field key, for which SetFunction is adopted
        :param kwargs: Function parameters
        :return: Modified params
        """
        if hstore_available():
            # Django before 1.10 doesn't convert HStoreField values to string automatically
            # Which causes a bug in cursor.execute(). Let's do it here
            from django.contrib.postgres.fields import HStoreField
            if isinstance(model._meta.get_field(key), HStoreField):
                kwargs[key] = hstore_serialize(kwargs[key])

        return kwargs

    def get_sql_value(self, field, val, connection, val_as_param=True, with_table=False, for_update=True, **kwargs):
        # type: (Field, Any, DefaultConnectionProxy, bool, bool, bool, **Any) -> Tuple[str, Tuple[Any]]
        """
        Returns value sql to set into field and parameters for query execution
        This method is called from get_sql() by default.
        :param field: Django field to take format from
        :param val: Value to format
        :param connection: Connection used to update data
        :param val_as_param: If flag is not set, value should be converted to string and inserted into query directly.
            Otherwise a placeholder and query parameter will be used
        :param with_table: If flag is set, column name in sql is prefixed by table name
        :param for_update: If flag is set, returns update sql. Otherwise - insert SQL
        :param kwargs: Additional arguments, if needed
        :return: A tuple: sql, replacing value in update and a tuple of parameters to pass to cursor
        """
        raise NotImplementedError("'%s' must define get_sql method" % self.__class__.__name__)

    def get_sql(self, field, val, connection, val_as_param=True, with_table=False, for_update=True, **kwargs):
        # type: (Field, Any, DefaultConnectionProxy, bool, bool, bool, **Any) -> Tuple[str, Tuple[Any]]
        """
        Returns function sql and parameters for query execution
        :param field: Django field to take format from
        :param val: Value to format
        :param connection: Connection used to update data
        :param val_as_param: If flag is not set, value should be converted to string and inserted into query directly.
            Otherwise a placeholder and query parameter will be used
        :param with_table: If flag is set, column name in sql is prefixed by table name
        :param for_update: If flag is set, returns update sql. Otherwise - insert SQL
        :param kwargs: Additional arguments, if needed
        :return: A tuple: sql, replacing value in update and a tuple of parameters to pass to cursor
        """
        val, params = self.get_sql_value(field, val, connection, val_as_param=val_as_param, with_table=with_table,
                                         for_update=for_update, **kwargs)
        return '"%s" = %s' % (field.column, val), params

    @classmethod
    def get_function_by_name(cls, name):  # type: (str) -> Optional[Type[AbstractSetFunction]]
        """
        Finds subclass of AbstractOperation applicable to given operation name
        :param name: String name to search
        :return: AbstractOperation subclass if found, None instead
        """
        try:
            return next(sub_cls for sub_cls in get_subclasses(cls, recursive=True) if name in sub_cls.names)
        except StopIteration:
            raise ValueError("Operation with name '%s' doesn't exist" % name)

    def field_is_supported(self, field):  # type: (Field) -> bool
        """
        Tests if this set function supports given field
        :param field: django.db.models.Field instance
        :return: Boolean
        """
        if self.supported_field_classes is None:
            return True
        else:
            return field.__class__.__name__ in self.supported_field_classes

    def _parse_null_default(self, field, connection, **kwargs):
        """
        Part of get_function_sql() method.
        When operation is done on the bases of field's previous value, we need a default to set instead of NULL.
        This function gets this default value.
        :param field: Field to set default for
        :param connection: Connection used to update data
        :param kwargs: kwargs of get_function_sql
        :return: null_default value
        """
        if 'null_default' in kwargs:
            null_default = kwargs['null_default']
        elif field.__class__.__name__ in NULL_DEFAULTS:
            null_default = NULL_DEFAULTS[field.__class__.__name__]
        else:
            raise Exception("Operation '%s' doesn't support field '%s'"
                            % (self.__class__.__name__, field.__class__.__name__))

        return self.format_field_value(field, null_default, connection)

    def _get_field_column(self, field, with_table=False):
        # type: (Field, bool) -> str
        """
        Returns quoted field column, prefixed with table name if needed
        :param field: Field instance
        :param with_table: Boolean flag - add table or not
        :return: String name
        """
        table = '"%s".' % field.model._meta.db_table if with_table else ''
        return '%s"%s"' % (table, field.column)


class EqualSetFunction(AbstractSetFunction):
    names = {'eq', '='}

    def get_sql_value(self, field, val, connection, val_as_param=True, with_table=False, for_update=True, **kwargs):
        if val_as_param:
            return self.format_field_value(field, val, connection)
        else:
            return '%s' % str(val), []


class EqualNotNullSetFunction(AbstractSetFunction):
    names = {'eq_not_null'}

    def modify_create_params(self, model, key, kwargs):
        if kwargs[key] is None:
            del kwargs[key]

        return kwargs

    def get_sql_value(self, field, val, connection, val_as_param=True, with_table=False, for_update=True, **kwargs):
        tpl = 'COALESCE(%s, %s)'
        if for_update:
            default_value, default_params = self._get_field_column(field, with_table=with_table), []
        else:
            default_value, default_params = self.format_field_value(field, field.get_default(), connection)
        if val_as_param:
            sql, params = self.format_field_value(field, val, connection)
            return tpl % (sql, default_value), params + default_params
        else:
            return tpl % (str(val), default_value), default_params


class PlusSetFunction(AbstractSetFunction):
    names = {'+', 'incr'}

    supported_field_classes = {'IntegerField', 'FloatField', 'AutoField', 'BigAutoField', 'BigIntegerField',
                               'SmallIntegerField', 'PositiveIntegerField', 'PositiveSmallIntegerField', 'DecimalField',
                               'IntegerRangeField', 'BigIntegerRangeField', 'FloatRangeField', 'DateTimeRangeField',
                               'DateRangeField'}

    def get_sql_value(self, field, val, connection, val_as_param=True, with_table=False, for_update=True, **kwargs):
        null_default, null_default_params = self._parse_null_default(field, connection, **kwargs)

        if val_as_param:
            sql, params = self.format_field_value(field, val, connection)
        else:
            sql, params = str(val), []

        if for_update:
            tpl = 'COALESCE(%s, %s) + %s'
            return tpl % (self._get_field_column(field, with_table=with_table), null_default, sql),\
                   null_default_params + params
        else:
            return sql, params


class ConcatSetFunction(AbstractSetFunction):
    names = {'||', 'concat'}

    supported_field_classes = {'CharField', 'TextField', 'EmailField', 'FilePathField', 'SlugField', 'HStoreField',
                               'URLField', 'BinaryField', 'JSONField', 'ArrayField', 'CITextField', 'CICharField',
                               'CIEmailField'}

    def get_sql_value(self, field, val, connection, val_as_param=True, with_table=False, for_update=True, **kwargs):
        null_default, null_default_params = self._parse_null_default(field, connection, **kwargs)

        # Postgres 9.4 has JSONB support, but doesn't support concat operator (||)
        # So I've taken function to solve the problem from
        # Note, that function should be created before using this operator
        if not for_update:
            tpl = '%s'
        elif get_postgres_version() < (9, 5) and isinstance(field, JSONField):
            tpl = '{0}(COALESCE(%s, %s), %s)'.format(Postgres94MergeJSONBMigration.FUNCTION_NAME)
        else:
            tpl = 'COALESCE(%s, %s) || %s'

        if val_as_param:
            val_sql, params = self.format_field_value(field, val, connection)
        else:
            val_sql, params = str(val), []

        if not for_update:
            return tpl % val_sql, params
        else:
            return tpl % (self._get_field_column(field, with_table=with_table), null_default, val_sql), null_default_params + params


class UnionSetFunction(AbstractSetFunction):
    names = {'union'}

    supported_field_classes = {'ArrayField'}

    def get_sql_value(self, field, val, connection, val_as_param=True, with_table=False, for_update=True, **kwargs):
        if for_update:
            sub_func = ConcatSetFunction()
            sql, params = sub_func.get_sql_value(field, val, connection, val_as_param=val_as_param,
                                                 with_table=with_table, **kwargs)
            sql = 'ARRAY(SELECT DISTINCT UNNEST(%s))' % sql
        else:
            sql, params = val, []

        return sql, params
